"""
This module contains the main disease modelling class.
"""
import copy
import logging
from datetime import datetime
from collections import OrderedDict
from typing import Callable, Dict, List, Optional, Tuple
from warnings import warn
import itertools

import networkx
import numpy as np
import pandas as pd

from computegraph import ComputeGraph

import summer2.flows as flows
from summer2.adjust import BaseAdjustment, FlowParam, Multiply
from summer2.compartment import Compartment
from summer2.derived_outputs import DerivedOutputRequest, calculate_derived_outputs
from summer2.parameters import params

from summer2.parameters.param_impl import finalize_parameters
from summer2.runner import ModelBackend
from summer2.solver import SolverType, solve_ode
from summer2.stratification import Stratification
from summer2.utils import get_scenario_start_index, ref_times_to_dti, clean_compartment_values
from summer2.population import get_unique_strat_groups, filter_by_strata
from summer2.tracker import ModelBuildTracker, ActionType

logger = logging.getLogger()

FlowRateFunction = Callable[[List[Compartment], np.ndarray, np.ndarray, float], np.ndarray]


class BackendType:
    PYTHON = "python"
    JAX = "jax"


class CompartmentalModel:
    """
    A compartmental disease model

    This model defines a set of compartments which each contain a population.
    Disease dynamics are defined by a set of flows which link the compartments together.
    The model is run over a period of time, starting from some initial conditions to
    predict the future state of a disease.

    Args:
        times: The start and end times. ***
        compartments: The compartments to simulate.
        infectious_compartments: The compartments which are counted as infectious.
        time_step (optional): The timesteps to return results for. This request does not
        affect the ODE solver, but is used for the stochastic solver. Defaults to ``1``.

    Attributes:
        times (np.ndarray): The times that the model will simulate.
        compartments (List[Compartment]): The model's compartments.
        initial_population (np.ndarray): The model's starting population. The indices of this
            array match to the ``compartments`` attribute. This is zero by default, but should
            be set with the ``set_initial_population`` method.
        outputs (np.ndarray): The values of each compartment for each requested timestep.
            For ``C`` compartments and ``T`` timesteps this will be a ``TxC`` matrix.
            The column indices of this array will match up with ``compartments`` and the
            row indices will match up with ``times``.
        derived_outputs (Dict[str, np.ndarray]): Additional results that are calculated from
            ``outputs`` for each timestep.


    """

    _DEFAULT_DISEASE_STRAIN = "default"
    _DEFAULT_MIXING_MATRIX = np.array([[1.0]])
    _DEFAULT_BACKEND = BackendType.JAX

    def __init__(
        self,
        times: Tuple[int, int],
        compartments: List[str],
        infectious_compartments: List[str],
        timestep: float = 1.0,
        ref_date: datetime = None,
    ):
        start_t, end_t = times
        assert end_t > start_t, "End time must be greater than start time"
        time_period = end_t - start_t
        num_steps = 1 + (time_period / timestep)
        msg = f"Time step {timestep} must be less than time period {time_period}"
        assert num_steps >= 1, msg
        msg = f"Time step {timestep} must be a factor of time period {time_period}"
        assert num_steps % 1 == 0, msg
        self.times = np.linspace(start_t, end_t, num=int(num_steps))
        self.timestep = timestep

        # Set the ref_date; the datetime object equivalent to times[0]
        self.ref_date = ref_date

        if isinstance(infectious_compartments, str):
            infectious_compartments = [infectious_compartments]
        error_msg = "Infectious compartments must be a subset of compartments"
        assert all(n in compartments for n in infectious_compartments), error_msg
        self.compartments = [Compartment(n, idx=i) for i, n in enumerate(compartments)]
        for c in self.compartments:
            if c in infectious_compartments:
                c.tags.append("infectious")

        # Compartment name lookup; needs to be present before adding any flows
        self._update_compartment_name_map()

        self._infectious_compartments = [Compartment(n) for n in infectious_compartments]
        self.initial_population = np.zeros_like(self.compartments, dtype=float)
        # Keeps track of original, pre-stratified compartment names.
        self._original_compartment_names = [Compartment.deserialize(n) for n in compartments]
        # Keeps track of Stratifications that have been applied.
        self._stratifications = []
        self.stratifications = {}
        # Flows to be applied to the model compartments
        self._flows = []

        # Turn on all runtime assertions by default; can be disabled for performance reasons
        # Setting to False still retains some checking, but turns off the most costly checks
        self.set_validation_enabled(True)
        self._derived_outputs_idx_cache = {}

        # The results calculated using the model: no outputs exist until the model has been run.
        self.outputs = None
        self.derived_outputs = None
        # Track 'derived output' requests in a dictionary.
        self._derived_output_requests = {}
        # Track 'derived output' request dependencies in a directed acylic graph (DAG).
        self._derived_output_graph = networkx.DiGraph()
        # Whitelist of 'derived outputs' to evaluate
        self._derived_outputs_whitelist = []

        # Map of (runtime) computed values
        self._computed_values_graph_dict = {}

        # Init baseline model to None; can be set via set_baseline if running as a scenario
        self._baseline = None

        # Mixing matrices: a list of square arrays, or functions, used to calculate
        # force of infection.
        self._mixing_matrices = []
        # Mixing categories: a list of dicts that knows the strata required to match a row in
        # the mixing matrix.
        self._mixing_categories = [{}]
        # Strains: a list of the different sub-categories ('strains') of the disease that we
        # are modelling.
        self._disease_strains = [self._DEFAULT_DISEASE_STRAIN]

        self._update_compartment_indices()

        # Track the actions we take building this model
        self.tracker = ModelBuildTracker()

        self._defer_actions = True

        self._finalized = False
        self._runner = None

    def _update_compartment_indices(self):
        """
        Update the mapping of compartment name to idx for quicker lookups.
        """
        compartment_idx_lookup = {}
        # Update the mapping of compartment name to idx for quicker lookups.
        for idx, c in enumerate(self.compartments):
            c.idx = idx
            compartment_idx_lookup[c] = idx

        for flow in self._flows:
            flow.update_compartment_indices(compartment_idx_lookup)

    def _assert_not_finalized(self):
        error_msg = "Cannot make changes to model that is already finalized"
        assert not self._finalized, error_msg

    def set_initial_population(self, distribution: Dict[str, float], force=False):
        """
        Sets the initial population of the model, which is zero by default.

        Args:
            distribution: A map of populations to be assigned to compartments.

        """
        if not force:
            self._assert_not_finalized()

            error_msg = "Cannot set initial population after the model has been stratified"
            assert not self._stratifications, error_msg

        assert isinstance(distribution, dict)

        # Make sure we're not supplying any eroneous compartment names
        for k, v in distribution.items():
            assert k in self._original_compartment_names

        self._init_pop_dist = distribution.copy()

        # Ensure dictionary contains all comparments (default to 0 if not supplied)
        for idx, comp in enumerate(self._original_compartment_names):
            if comp not in self._init_pop_dist:
                self._init_pop_dist[comp] = 0

    def finalize(self):
        if not self._finalized:
            finalize_parameters(self)
            self._finalized = True

    def get_object_graph(self, obj) -> ComputeGraph:
        """Return the ComputeGraph for this object

        Args:
            obj: Any summer object with a graph_key
        """
        if isinstance(obj, dict):
            targets = [v._graph_key for v in obj.values()]
        else:
            targets = obj._graph_key
        return self.graph.filter(targets)

    def set_validation_enabled(self, validate: bool):
        """
        Set this to False in order to turn of some (potentially expensive) runtime validation
        E.g. In calibration, leave it enabled for the first iteration, to catch any structural
        model issues, but then disable for subsequent iterations
        """
        self._should_validate = validate

    def _set_derived_outputs_index_cache(self, idx_cache: dict):
        self._derived_outputs_idx_cache = idx_cache

    """
    Adding flows
    """

    def add_crude_birth_flow(
        self,
        name: str,
        birth_rate: FlowParam,
        dest: str,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a crude birth rate flow to the model.
        The number of births will be determined by the product of the birth rate
        and total population.

        Args:
            name: The name of the new flow.
            birth_rate: The fractional crude birth rate per timestep.
            dest: The name of the destination compartment.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows
                                            are created.

        """
        is_already_birth_flow = any(
            [
                type(f) is flows.CrudeBirthFlow or type(f) is flows.ReplacementBirthFlow
                for f in self._flows
            ]
        )
        if is_already_birth_flow:
            msg = "There is already a birth flow in this model, cannot add a second."
            raise ValueError(msg)

        self._add_entry_flow(
            flows.CrudeBirthFlow,
            name,
            birth_rate,
            dest,
            dest_strata,
            expected_flow_count,
        )

    def add_replacement_birth_flow(
        self,
        name: str,
        dest: str,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a death-replacing birth flow to the model.
        The number of births will replace the total number of deaths each year

        Args:
            name: The name of the new flow.
            dest: The name of the destination compartment.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows
                                            are created.

        """
        # Only allow a single replacement flow to be added to the model.
        is_already_birth_flow = any(
            [
                type(f) is flows.CrudeBirthFlow or type(f) is flows.ReplacementBirthFlow
                for f in self._flows
            ]
        )
        if is_already_birth_flow:
            msg = "There is already a birth flow in this model, cannot add a second."
            raise ValueError(msg)

        self._add_entry_flow(
            flows.ReplacementBirthFlow,
            name,
            1.0,
            dest,
            dest_strata,
            expected_flow_count,
        )

    def add_importation_flow(
        self,
        name: str,
        num_imported: FlowParam,
        dest: str,
        split_imports: bool,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds an importation flow to the model, where people enter the destination compartment from
        outside the system.
        The number of people imported per timestep is completely determined by ``num_imported``.

        Args:
            name: The name of the new flow.
            num_imported: The number of people arriving per timestep.
            dest: The name of the destination compartment.
            split_imports: Whether to split num_imported amongst existing destination
                           compartments (True), or add the full value to each (False)
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are
                                            created.

        """

        dest_strata = dest_strata or {}
        dest_comps = [c for c in self.compartments if c.is_match(dest, dest_strata)]

        if split_imports:
            # Adjust each of the flows so they are split equally between dest_comps
            adjustments = [Multiply(1.0 / len(dest_comps))]
        else:
            # No adjustment - flow to each dest will be num_imported
            adjustments = None

        self._add_entry_flow(
            flows.ImportFlow,
            name,
            num_imported,
            dest,
            dest_strata,
            expected_flow_count,
            adjustments,
        )

    def _add_entry_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        dest: str,
        dest_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
        adjustments: List[BaseAdjustment] = None,
    ):
        self._assert_not_finalized()
        dest_strata = dest_strata or {}
        dest_comps = [c for c in self.compartments if c.is_match(dest, dest_strata)]
        new_flows = []
        for dest_comp in dest_comps:
            flow = flow_cls(name, dest_comp, param, adjustments=adjustments)
            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows

    def add_death_flow(
        self,
        name: str,
        death_rate: FlowParam,
        source: str,
        source_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow where people die and leave the compartment, reducing the total population.

        Args:
            name: The name of the new flow.
            death_rate: The fractional death rate per timestep.
            source: The name of the source compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are
                                            created.

        """
        self._add_exit_flow(
            flows.DeathFlow,
            name,
            death_rate,
            source,
            source_strata,
            expected_flow_count,
        )

    def add_universal_death_flows(self, base_name: str, death_rate: FlowParam):
        """
        Adds a universal death rate flow to every compartment in the model.
        The number of deaths per compartment will be determined by the product of
        the death rate and the compartment population.

        The base name will be used to create the name of each flow. For example a
        base name of "universal_death" applied to the "S" compartment will result in a flow called
        "universal_death_for_S".

        Args:
            base_name: The base name for each new flow.
            death_rate: The fractional death rate per timestep.

        Returns:
            List[str]: The names of the flows added.

        """

        # Only allow a single universal death flow with a given name to be added to the model.
        is_already_used = any([f.name.startswith(base_name) for f in self._flows])
        if is_already_used:
            msg = f"There is already a universal death flow called '{base_name}' in this model, \
                cannot add a second."
            raise ValueError(msg)

        flow_names = []
        for comp_name in self._original_compartment_names:
            flow_name = f"{base_name}_for_{comp_name}"
            flow_names.append(flow_name)
            self._add_exit_flow(
                flows.DeathFlow,
                flow_name,
                death_rate,
                comp_name,
                source_strata={},
                expected_flow_count=None,
            )

        return flow_names

    def _add_exit_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        source: str,
        source_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
    ):
        self._assert_not_finalized()
        source_strata = source_strata or {}
        source_comps = [c for c in self.compartments if c.is_match(source, source_strata)]
        new_flows = []
        for source_comp in source_comps:
            flow = flow_cls(name, source_comp, param)
            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows

    def add_infection_frequency_flow(
        self,
        name: str,
        contact_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow that infects people using an "infection frequency" contact rate, which is
        when the infectious multiplier is determined by the proportion of infectious people to the
        total population.

        Args:
            name: The name of the new flow.
            contact_rate: The effective contact rate.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are
                                            created.

        """
        self._add_transition_flow(
            flows.InfectionFrequencyFlow,
            name,
            contact_rate,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
            find_infectious_multiplier=self._get_infection_frequency_multiplier,
        )

    def add_infection_density_flow(
        self,
        name: str,
        contact_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
    ):
        """
        Adds a flow that infects people using an "infection density" contact rate, which is
        when the infectious multiplier is determined by the number of infectious people.

        Args:
            name: The name of the new flow.
            contact_rate: The contact rate.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are
                                            created.

        """
        self._add_transition_flow(
            flows.InfectionDensityFlow,
            name,
            contact_rate,
            source,
            dest,
            source_strata,
            dest_strata,
            expected_flow_count,
            find_infectious_multiplier=self._get_infection_density_multiplier,
        )

    def add_transition_flow(
        self,
        name: str,
        fractional_rate: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        expected_flow_count: Optional[int] = None,
        absolute=False,
    ):
        """
        Adds a flow transferring people from a source to a destination based on the population of
        the source compartment and the fractional flow rate.

        Args:
            name: The name of the new flow.
            fractional_rate: The fraction of people that transfer per timestep.
            source: The name of the source compartment.
            dest: The name of the destination compartment.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            expected_flow_count (optional): Used to assert that a particular number of flows are
                                            created.

        """
        if absolute:
            self._add_transition_flow(
                flows.AbsoluteFlow,
                name,
                fractional_rate,
                source,
                dest,
                source_strata,
                dest_strata,
                expected_flow_count,
            )
        else:
            self._add_transition_flow(
                flows.TransitionFlow,
                name,
                fractional_rate,
                source,
                dest,
                source_strata,
                dest_strata,
                expected_flow_count,
            )

    def _add_transition_flow(
        self,
        flow_cls,
        name: str,
        param: FlowParam,
        source: str,
        dest: str,
        source_strata: Optional[Dict[str, str]],
        dest_strata: Optional[Dict[str, str]],
        expected_flow_count: Optional[int],
        find_infectious_multiplier=None,
    ):
        self._assert_not_finalized()
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}

        dest_comps = self.get_matching_compartments(dest, dest_strata)
        source_comps = self.get_matching_compartments(source, source_strata)

        num_dest = len(dest_comps)
        num_source = len(source_comps)
        msg = f"Expected equal number of source and dest compartments, but got {num_source} source \
                and {num_dest} dest."
        assert num_dest == num_source, msg
        new_flows = []
        for source_comp, dest_comp in zip(source_comps, dest_comps):
            if find_infectious_multiplier:
                flow = flow_cls(
                    name,
                    source_comp,
                    dest_comp,
                    param,
                    find_infectious_multiplier=find_infectious_multiplier,
                )
            else:
                flow = flow_cls(name, source_comp, dest_comp, param)

            new_flows.append(flow)

        self._validate_expected_flow_count(expected_flow_count, new_flows)
        self._flows += new_flows

    @staticmethod
    def _validate_expected_flow_count(
        expected_count: Optional[int], new_flows: List[flows.BaseFlow]
    ):
        """
        Ensure the number of new flows created is the expected amount

        """
        if expected_count is not None:
            # Check that we added the expected number of flows.
            actual_count = len(new_flows)
            msg = f"Expected to add {expected_count} flows but added {actual_count}"
            assert actual_count == expected_count, msg

    """
    Methods for searching and indexing flows and compartments
    """

    def _update_compartment_name_map(self):
        names = set([c.name for c in self.compartments])
        name_map = {}
        for n in names:
            name_map[n] = [c for c in self.compartments if c.name == n]
        self._compartment_name_map = name_map

    def get_matching_compartments(self, name: str, strata: dict):
        if isinstance(name, str):
            name_query = self._compartment_name_map[name]
        else:
            # FIXME: Should do better type checking here
            # For now we assume we have some kind of iterable (ie a 'list' of names)
            match_lists = [self._compartment_name_map[n] for n in name]
            name_query = list(itertools.chain.from_iterable(match_lists))

        if not len(strata):
            return name_query
        else:
            _strata = frozenset(strata.items())
            return [c for c in name_query if c._has_strata(_strata)]

    def _get_matching_compartments(self, name: str, strata: frozenset):
        """
        Optimized version of above method that uses frozenset directly
        """
        name_query = self._compartment_name_map[name]

        if not len(strata):
            return name_query
        else:
            return [c for c in name_query if c._has_strata(strata)]

    """
    Stratifying the model
    """

    def stratify_with(self, strat: Stratification):
        """
        Apply the stratification to the model's flows and compartments.

        Args:
            strat: The stratification to apply.

        """

        self._assert_not_finalized()
        # Enable/disable runtime assertions for strat
        strat._validate = self._should_validate

        # Validate flow adjustments
        flow_names = [f.name for f in self._flows]
        for n in strat.flow_adjustments.keys():
            msg = f"Flow adjustment for '{n}' refers to a flow that is not present in the model."
            assert n in flow_names, msg

        for fadj in strat.flow_adjustments.values():
            for _, source_strata, dest_strata in fadj:
                # Verify source
                self._strata_exist(source_strata)
                # Veryify dest
                self._strata_exist(dest_strata)

        # Validate infectiousness adjustments.
        msg = "All stratification infectiousness adjustments must refer to a compartment that is \
               present in model."
        assert all(
            [c in self._original_compartment_names for c in strat.infectiousness_adjustments.keys()]
        ), msg

        if strat.mixing_matrix is not None:
            # Add this stratification's mixing matrix if a new one is provided.
            assert not strat.is_strain(), "Strains cannot have a mixing matrix."
            # Only allow mixing matrix to be supplied if there is a complete stratification.
            msg = "Mixing matrices only allowed for full stratification."
            assert strat.compartments == self._original_compartment_names, msg
            self._mixing_matrices.append(strat.mixing_matrix)
            # Update mixing categories for force of infection calculation.
            old_mixing_categories = self._mixing_categories
            self._mixing_categories = []
            for mc in old_mixing_categories:
                for stratum in strat.strata:
                    self._mixing_categories.append({**mc, strat.name: stratum})

        if strat.is_strain():
            # Track disease strain names, overriding default values.
            msg = "An infection strain stratification has already been applied, cannot use this \
                   more than once."
            assert not any([s.is_strain() for s in self._stratifications]), msg
            self._disease_strains = strat.strata

        for c in strat.compartments:
            if c not in self._original_compartment_names:
                raise Exception("Trying to stratify non-existent compartment", c)
        # Stratify compartments, split according to split_proportions
        prev_compartment_names = copy.copy(self.compartments)
        self.compartments = strat._stratify_compartments(self.compartments)
        # if not self._defer_actions or self._finalized:
        #    try:
        #        self.initial_population = strat._stratify_compartment_values(
        #            prev_compartment_names, self.initial_population
        #        )
        #    except Exception as e:
        #        logger.critical("Parameterized models must set takes_params in constructor")
        #        raise e

        # Update the cache of compartment names; these need to correct whenever we add a new flow
        self._update_compartment_name_map()

        # Stratify flows
        prev_flows = self._flows
        self._flows = []
        for flow in prev_flows:
            self._flows += flow.stratify(strat)

        # Update the mapping of compartment name to idx for quicker lookups.
        self._update_compartment_indices()

        if strat.is_ageing():
            msg = "Age stratification can only be applied once"
            assert not any([s.is_ageing() for s in self._stratifications]), msg

            # Create inter-compartmental flows for ageing from one stratum to the next.
            # The ageing rate is proportional to the width of the age bracket.
            # It's assumed that both ages and model timesteps are in years.
            ages = list(sorted(map(int, strat.strata)))

            # Only allow age stratification to be applied with complete stratifications,
            # because everyone has an age.
            msg = "Mixing matrices only allowed for full stratification."
            assert strat.compartments == self._original_compartment_names, msg

            for age_idx in range(len(ages) - 1):
                start_age = int(ages[age_idx])
                end_age = int(ages[age_idx + 1])
                for comp in prev_compartment_names:

                    source = comp.stratify(strat.name, str(start_age))
                    dest = comp.stratify(strat.name, str(end_age))
                    ageing_rate = 1.0 / (end_age - start_age)
                    self.add_transition_flow(
                        name=f"ageing_{source}_to_{dest}",
                        fractional_rate=ageing_rate,
                        source=source.name,
                        dest=dest.name,
                        source_strata=source.strata,
                        dest_strata=dest.strata,
                        expected_flow_count=1,
                    )

        self._stratifications.append(strat)
        self.stratifications[strat.name] = strat

        self.tracker.append_action(ActionType.STRATIFY, strat=strat)

    def _strata_exist(self, strata: dict):
        """
        Verify whether all the strata exist within the model
        Raises an Exception if not
        """
        strat_names = [s.name for s in self._stratifications]
        for k, v in strata.items():
            if k not in strat_names:
                raise KeyError(f"Invalid stratification {k}")
            for s in self._stratifications:
                if k == s.name:
                    if v not in s.strata:
                        raise ValueError(f"Invalid stratum {v} for {s}")

    def get_stratification(self, name: str) -> Stratification:
        """Return Stratification matching name, or None if not found

        Args:
            name: Name of stratificaton to get

        Returns:
            Stratification
        """
        for stratification in self._stratifications:
            if stratification.name == name:
                return stratification
        return None

    def adjust_population_split(self, strat: str, dest_filter: dict, proportions: dict):
        """Adjust the initial population to redistribute the population for a particular
        stratification, over a subset of some other strata

        Args:
            strat (str): The stratification to redistribute over
            dest_filter (dict): Subset of (other) strata to filter the split by
            proportions (dict): Proportions of new split (must have all strata specified)

        """
        self._assert_not_finalized()

        msg = f"No stratification {strat} found in model"
        assert strat in [s.name for s in self._stratifications], msg

        model_strat = [s for s in self._stratifications if s.name == strat][0]

        msg = "All strata must be specified in proportions"
        assert set(model_strat.strata) == set(proportions), msg

        msg = "Proportions must sum to 1.0"
        np.testing.assert_allclose(sum(proportions.values()), 1.0, err_msg=msg)

        # strat_comps = [c for c in self.compartments if strat in c.strata]
        # # Filter by only the subset we're setting in split_map
        # strat_comps = filter_by_strata(strat_comps, dest_filter)

        # usg = get_unique_strat_groups(strat_comps, strat)

        # for g in usg:
        #     mcomps = self._get_matching_compartments(g.name, g.strata)
        #     idx = [c.idx for c in mcomps]
        #     total = self.initial_population[idx].sum()
        #     for c in mcomps:
        #         k = c.strata[strat]
        #         target_prop = proportions[k]
        #         self.initial_population[c.idx] = total * target_prop

        self.tracker.append_action(
            ActionType.ADJUST_POP_SPLIT,
            strat=strat,
            dest_filter=dest_filter,
            proportions=proportions,
        )

    """
    Running the model
    """

    def _get_step_test(self, parameters: dict = None):
        self._update_compartment_indices()
        self.finalize()

        self._set_backend("jax")
        # self._backend.prepare_to_run(parameters)
        self._backend.prepare_structural()

        from summer2.runner.jax.model_impl import build_run_model

        jax_run_func, jax_runner_dict = build_run_model(self._backend, base_params=parameters)

        return jax_runner_dict["one_step"](parameters)

    def get_runner(self, parameters: dict, dyn_params: List = None, jit=True, **backend_args):
        self._update_compartment_indices()
        self.finalize()

        self._set_backend("jax")
        # self._backend.prepare_to_run(parameters)
        self._backend.prepare_structural()

        from summer2.runner.jax.model_impl import build_run_model

        jax_run_func, jax_runner_dict = build_run_model(
            self._backend, base_params=parameters, dyn_params=dyn_params, **backend_args
        )
        if jit:
            from jax import jit as jjit

            jax_run_func = jjit(jax_run_func)

        return ModelResults(self, jax_run_func, jax_runner_dict)

    def run(
        self,
        parameters: dict = None,
        solver: str = SolverType.SOLVE_IVP,
        backend_args: dict = None,
        **kwargs,
    ):
        """
        Runs the model over the provided time span, calculating the outputs and the derived outputs.
        The model calculates the outputs by solving an ODE which is defined by the initial
        population and the inter-compartmental flows.

        **Note**: The default ODE solver used to produce the model's outputs does not necessarily
        evaluate every requested timestep. This adaptive
        solver can skip over times, or double back when trying to characterize the ODE. The final
        results are produced by interpolating the
        solution produced by the ODE solver. This means that model dynamics that only occur in
        short time periods may not be reflected in the outputs.

        Args:
            solver (optional): The ODE solver to use, defaults to SciPy's IVP solver.
            **kwargs (optional): Extra arguments to supplied to the solver, see ``summer.solver``
            for details.

        """

        # Ensure we call this before model runs, since it is now disabled inside individual
        # flow constructors
        self._update_compartment_indices()
        self.finalize()

        parameters = parameters or {}

        parameters = {k: v for k, v in parameters.items() if k in self.get_input_parameters()}

        if self._runner is None:
            self._set_backend("jax", backend_args)
            self._backend.prepare_structural()
            self._runner = self.get_runner(parameters, solver=solver, **kwargs)

        self._runner.run(parameters=parameters)

    def _set_backend(self, backend: str, backend_args: dict = None):
        backend_args = backend_args or {}
        self._backend = ModelBackend(self, **backend_args)

    def _solve_ode(self, solver, solver_args: dict):
        """
        Runs the model over the provided time span, calculating the outputs.
        Uses an ODE interpretation of flow rates.
        """

        # Calculate the outputs (compartment sizes) by solving the ODE defined by
        # _get_compartment_rates().
        self.outputs = solve_ode(
            solver,
            self._backend.get_compartment_rates,
            self.initial_population,
            self.times,
            solver_args,
        )

    def _get_infection_frequency_multiplier(self, *args, **kwargs) -> float:
        return self._backend._get_infection_frequency_multiplier(*args, **kwargs)

    def _get_infection_density_multiplier(self, *args, **kwargs):
        return self._backend._get_infection_density_multiplier(*args, **kwargs)

    def _get_timestep_deaths(self, *args, **kwargs):
        return self._backend._get_timestep_deaths(*args, **kwargs)

    """
    Requesting and calculating derived outputs
    """

    def set_derived_outputs_whitelist(self, whitelist: List[str]):
        """
        Request that we should only calculate a subset of the model's derived outputs.
        This can be useful when you only care about some results and you want to cut down on
        runtime.
        For example, we may only need some derived outputs for calibration, but may need more later
        when we want to know all the dyanmics that the model actually showed.

        Args:
            whitelist: A list of the derived output names to calculate, ignoring all others.

        """
        self._derived_outputs_whitelist = whitelist

    def set_baseline(self, baseline):
        """Set a baseline model to be used for this (scenario) run
        Sets initial population values to the baseline values for this model's start time
        Cumulative and relative outputs will refer to the baseline

        Args:
            baseline (CompartmentalModel): The baseline model to be used as reference
        """
        start_index = get_scenario_start_index(baseline.times, self.times[0])
        init_compartments = baseline.outputs[start_index, :]
        self.initial_population = init_compartments
        self._baseline = baseline

    def _calculate_derived_outputs(self):
        return calculate_derived_outputs(
            requests=self._derived_output_requests,
            graph=self._derived_output_graph,
            outputs=self.outputs,
            times=self.times,
            timestep=self.timestep,
            flows=self._flows,
            compartments=self.compartments,
            get_flow_rates=self._backend.get_flow_rates,
            model=self,
            whitelist=self._derived_outputs_whitelist,
            baseline=self._baseline,
            idx_cache=self._derived_outputs_idx_cache,
            parameters=self._backend.parameters,
        )

    def request_output_for_flow(
        self,
        name: str,
        flow_name: str,
        source_strata: Optional[Dict[str, str]] = None,
        dest_strata: Optional[Dict[str, str]] = None,
        save_results: bool = True,
        raw_results: bool = False,
    ):
        """
        Adds a derived output to the model's results.
        The output will be the value of the requested flow at each timestep.

        Args:
            name: The name of the derived output.
            flow_name: The name of the flow to track.
            source_strata (optional): A whitelist of strata to filter the source compartments.
            dest_strata (optional): A whitelist of strata to filter the destination compartments.
            save_results (optional): Whether to save or discard the results. Defaults to ``True``.
            raw_results (optional): Whether to use raw interpolated flow rates, or post-process them
                so that they're more representative of the changes in compartment sizes.
                Defaults to ``False``.
        """
        source_strata = source_strata or {}
        dest_strata = dest_strata or {}

        if self._should_validate:
            msg = f"A derived output named {name} already exists."
            assert name not in self._derived_output_requests, msg
            is_flow_exists = any(
                [f.is_match(flow_name, source_strata, dest_strata) for f in self._flows]
            )
            assert is_flow_exists, f"No flow matches: {flow_name} {source_strata} {dest_strata}"

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.FLOW,
            "flow_name": flow_name,
            "source_strata": source_strata,
            "dest_strata": dest_strata,
            "raw_results": raw_results,
            "save_results": save_results,
        }

    def request_output_for_compartments(
        self,
        name: str,
        compartments: List[str],
        strata: Optional[Dict[str, str]] = None,
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output
        will be the aggregate population of the requested compartments at the at each timestep.

        Args:
            name: The name of the derived output.
            compartments: The name of the compartments to track.
            strata (optional): A whitelist of strata to filter the compartments.
            save_results (optional): Whether to save or discard the results.
        """
        strata = strata or {}

        if isinstance(compartments, str):
            compartments = [compartments]

        if self._should_validate:
            msg = f"A derived output named {name} already exists."
            assert name not in self._derived_output_requests, msg
            is_match_exists = any(
                [
                    any([c.is_match(name, strata) for name in compartments])
                    for c in self.compartments
                ]
            )
            assert is_match_exists, f"No compartment matches: {compartments} {strata}"

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.COMPARTMENT,
            "compartments": compartments,
            "strata": strata,
            "save_results": save_results,
        }

    def request_aggregate_output(
        self,
        name: str,
        sources: List[str],
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the aggregate of other
        derived outputs.

        Args:
            name: The name of the derived output.
            sources: The names of the derived outputs to aggregate.
            save_results (optional): Whether to save or discard the results.

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        for source in sources:
            assert (
                source in self._derived_output_requests
            ), f"Source {source} has not been requested."
            self._derived_output_graph.add_edge(source, name)

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.AGGREGATE,
            "sources": sources,
            "save_results": save_results,
        }

    def request_cumulative_output(
        self,
        name: str,
        source: str,
        start_time: int = None,
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the accumulated values of
        another derived output over the model's time period.

        Args:
            name: The name of the derived output.
            source: The name of the derived outputs to accumulate.
            start_time (optional): The time to start accumulating from, defaults to model start time
            save_results (optional): Whether to save or discard the results.

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        assert source in self._derived_output_requests, f"Source {source} has not been requested."
        self._derived_output_graph.add_node(name)
        self._derived_output_graph.add_edge(source, name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.CUMULATIVE,
            "source": source,
            "start_time": start_time,
            "save_results": save_results,
        }

    def request_function_output(
        self,
        name: str,
        func: Callable[[np.ndarray], np.ndarray],
        sources: List[str],
        save_results: bool = True,
    ):
        """
        Adds a derived output to the model's results. The output will be the result of a function
        which takes a list of sources as an input.

        Args:
            name: The name of the derived output.
            func: A function used to calculate the derived ouput.
            sources: The derived outputs to input into the function.
            save_results (optional): Whether to save or discard the results.

        Example:
            Request a function-based derived output:

                model.request_output_for_compartments(
                    compartments=["S", "E", "I", "R"],
                    name="total_population",
                    save_results=False
                )
                model.request_output_for_compartments(
                    compartments=["R"],
                    name="recovered_population",
                    save_results=False
                )

                def calculate_proportion_seropositive(recovered_pop, total_pop):
                    return recovered_pop / total_pop

                model.request_function_output(
                    name="proportion_seropositive",
                    func=calculate_proportion_seropositive,
                    sources=["recovered_population", "total_population"],
                )

        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        for source in sources:
            assert (
                source in self._derived_output_requests
            ), f"Source {source} has not been requested."
            self._derived_output_graph.add_edge(source, name)

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.FUNCTION,
            "func": func,
            "sources": sources,
            "save_results": save_results,
        }

    def request_param_function_output(
        self, name: str, func: params.Function, save_results: bool = True
    ):
        """Request a generic Function output

        Args:
            name (str): _description_
            func (Function): The Function, whose args are Param
            save_results (bool, optional): _description_. Defaults to True.
        """

        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg
        for k, v in func.kwargs.items():
            if isinstance(v, params.DerivedOutput):
                source = v.key
                assert (
                    source in self._derived_output_requests
                ), f"Source {source} has not been requested."
                self._derived_output_graph.add_edge(source, name)

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.PARAM_FUNCTION,
            "func": func,
            "save_results": save_results,
        }

    def request_computed_value_output(self, name: str, save_results: bool = True):
        """
        Save a computed value process output to derived outputs

        Args:
            name (str): Name (key) of computed value process
            save_results (bool, optional): Save outputs (or discard if False)
        """
        msg = f"A derived output named {name} already exists."
        assert name not in self._derived_output_requests, msg

        self._derived_output_graph.add_node(name)
        self._derived_output_requests[name] = {
            "request_type": DerivedOutputRequest.COMPUTED_VALUE,
            "name": name,
            "save_results": save_results,
        }

    def add_computed_value_process(self, name: str, processor):
        """
        Calculate (at runtime) values derived from the current compartment values and/or
        functions/input data
        providing time varying shared values.  The output values of these processes can be used by
        function parameters, adjustments, and function flows.

        Args:
            name (str): Name (key) of derived value (use this when referencing it in functions)
            processor (DerivedValueProcessor): Object providing implementation
        """
        # FIXME: We might actually have to keep this for now, at least until modellers get sick of
        # seeing it and change over all the code
        raise DeprecationWarning("Use add_computed_value_func instead")

    def add_computed_value_func(self, name: str, func: params.Function):
        if name in self._computed_values_graph_dict:
            raise Exception(f"Computed value function with name {name} already exists")
        self._computed_values_graph_dict[name] = func

    def get_computed_value_keys(self):
        return list(self._computed_values_graph_dict)

    def _get_ref_idx(self):
        if self.ref_date:
            times = ref_times_to_dti(self.ref_date, self.times)
        else:
            times = self.times
        return times

    def get_outputs_df(self):
        idx = self._get_ref_idx()
        column_str = [str(c) for c in self.compartments]
        return pd.DataFrame(self.outputs, index=idx, columns=column_str)

    def get_derived_outputs_df(self):
        idx = self._get_ref_idx()
        return pd.DataFrame(self.derived_outputs, index=idx)

    def get_input_parameters(self):
        if not self._finalized:
            raise Exception("Cannot trace input parameters before model is finalized")
        all_in_var = set(self.graph.get_input_variables())
        all_in_var = all_in_var.union(set(self._do_tracker_graph.get_input_variables()))
        return set([v.key for v in all_in_var if v.source == "parameters"])

    def query_compartments(self, query: dict = None, tags: List = None, as_idx=False):
        from summer2.inspect import query_compartments

        return query_compartments(self, query, tags, as_idx)

    def query_flows(
        self,
        flow_name: str = None,
        source: dict = None,
        dest: dict = None,
        tags: List = None,
    ):
        from summer2.inspect import query_flows

        return query_flows(self, flow_name, source, dest, tags)


class ModelResults:
    def __init__(self, model, run_func, runner_dict=None):
        self.model = model
        self._run_func = run_func
        self._input_params = model.get_input_parameters()
        self._derived_outputs_idx_cache = None
        self._runner_dict = runner_dict

    def get_outputs_df(self):
        return self.model.get_outputs_df()

    def get_derived_outputs_df(self):
        return self.model.get_derived_outputs_df()

    def run(self, parameters: dict, filter=True):
        if filter:
            parameters = {k: v for k, v in parameters.items() if k in self._input_params}

        results = self._run_func(parameters=parameters)

        self.outputs = np.array(results["outputs"])
        self.derived_outputs = {k: np.array(v) for k, v in results["derived_outputs"].items()}
        self.model.outputs = self.outputs
        self.model.derived_outputs = self.derived_outputs
        return results



"""
General utility functions used accross submodules
"""

import pandas as pd
from datetime import timedelta, datetime
from typing import Iterable
import numpy as np


def get_scenario_start_index(base_times, start_time):
    """
    Returns the index of the closest time step that is at, or before the scenario start time.
    """
    assert (
        base_times[0] <= start_time
    ), f"Scenario start time {start_time} is before baseline has started"
    indices_after_start_index = [idx for idx, time in enumerate(base_times) if time > start_time]
    if not indices_after_start_index:
        raise ValueError(f"Scenario start time {start_time} is set after the baseline time range")

    index_after_start_index = min(indices_after_start_index)
    start_index = max([0, index_after_start_index - 1])
    return start_index


def ref_times_to_dti(ref_date: datetime, times: Iterable) -> pd.DatetimeIndex:
    """Return a DatetimeIndex whose values after ref_date offset by times

    Args:
        ref_date (datetime): The reference  date ('epoch') from which offsets are computed
        times (Iterable): Offsets to ref_date (int or float)

    Returns:
        (pd.DatetimeIndex): Index suitable for constructing Series or DataFrame
    """

    return pd.DatetimeIndex([ref_date + timedelta(t) for t in times])


def clean_compartment_values(compartment_values: np.ndarray):
    """
    Zero out -ve compartment sizes in flow rate calculations,
    to prevent negative values from messing up the direction of flows.
    We don't expect large -ve values, but there can be small ones due to numerical errors.
    """
    comp_vals = compartment_values.copy()
    zero_mask = comp_vals < 0
    comp_vals[zero_mask] = 0
    return comp_vals



# Reimplementation of stratification and initial pop calculation for Jax runners

from __future__ import annotations
from typing import List, TYPE_CHECKING

from jax import numpy as jnp

from summer2.parameters import get_static_param_value, is_var, Function
from summer2.parameters.param_impl import ModelParameter
from summer2.stratification import Stratification

from summer2.population import get_rebalanced_population

if TYPE_CHECKING:
    from summer2 import CompartmentalModel


def get_stratify_compartments_func(
    model: CompartmentalModel, strat: Stratification, input_comps: List[str]
):
    """Build the equivalent of self._stratify_compartment_values for a given Stratification

    Args:
        strat (Stratification): The stratification to build this for
        input_comps (List[str]): The list of input compartments (pre-stratification)
    """

    def _stratify_compartment_values(comp_values: jnp.ndarray, static_graph_values: dict = None):
        """
        Stratify the model compartments into sub-compartments, based on the strata names provided.
        Split the population according to the provided proportions.
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartment values.
        """
        new_comp_values = []

        population_split = get_static_param_value(strat.population_split, static_graph_values)

        for idx in range(len(comp_values)):
            should_stratify = input_comps[idx].has_name_in_list(strat.compartments)
            if should_stratify:
                for stratum in strat.strata:
                    new_value = comp_values[idx] * population_split[stratum]
                    new_comp_values.append(new_value)
            else:
                new_comp_values.append(comp_values[idx])

        return jnp.array(new_comp_values)

    def stratify_compartment_values(comp_values: jnp.ndarray, static_graph_values: dict = None):
        """
        Stratify the model compartments into sub-compartments, based on the strata names provided.
        Split the population according to the provided proportions.
        Only compartments specified in the stratification's definition will be stratified.
        Returns the new compartment values.
        """
        population_split = get_static_param_value(strat.population_split, static_graph_values)

        new_comp_values = jnp.empty(strat._new_size)
        new_comp_values = new_comp_values.at[strat._passthrough_target_indices].set(
            comp_values[strat._passthrough_base_indices]
        )

        base_values = comp_values[strat._strat_base_indices]
        for stratum in strat.strata:
            new_value = base_values * population_split[stratum]
            new_comp_values = new_comp_values.at[strat._stratum_target_indices[stratum]].set(
                new_value
            )
        return new_comp_values

    return stratify_compartment_values


def get_calculate_initial_pop(model: CompartmentalModel):
    """This is the function we're most interested in

    Args:
        model (_type_): _description_

    Raises:
        NotImplementedError: _description_
        TypeError: _description_

    Returns:
        _type_: _description_
    """

    strat_funcs = {}
    comps = model._original_compartment_names
    for strat in model._stratifications:
        strat_funcs[strat] = get_stratify_compartments_func(model, strat, comps)
        comps = strat._stratify_compartments(comps)

    def calculate_initial_population(static_graph_values: dict) -> jnp.ndarray:
        """
        Called to recalculate the initial population from either fixed dictionary, or a dict
        supplied as a parameter
        """
        # FIXME:
        # Work in progress; correctly recalculates non-parameterized
        # populations, but does not include population rebalances etc
        distribution = model._init_pop_dist
        initial_population = jnp.zeros(len(model._original_compartment_names))

        if isinstance(distribution, dict):
            for idx, comp in enumerate(model._original_compartment_names):
                pop = get_static_param_value(distribution[comp.name], static_graph_values)
                initial_population = initial_population.at[idx].set(pop)

            for action in model.tracker.all_actions:
                if action.action_type == "stratify":
                    strat = action.kwargs["strat"]
                    initial_population = strat_funcs[strat](initial_population, static_graph_values)
                elif action.action_type == "adjust_pop_split":
                    # FIXME: Implement this
                    initial_population = get_rebalanced_population(
                        model, initial_population, static_graph_values, **action.kwargs
                    )
            return initial_population
        else:
            raise TypeError(
                "Initial population distribution must be a dict or a Function that returns one",
                distribution,
            )

    return calculate_initial_population



"""
Tools for solving compartmental ODEs
"""
from typing import Callable

import numpy as np
from scipy.integrate import odeint, solve_ivp
from scipy.interpolate import interp1d

OdeFunction = Callable[[np.ndarray, float], np.ndarray]


class SolverType:
    """
    Options for ODE solver used by model
    """

    ODE_INT = "odeint"
    SOLVE_IVP = "solve_ivp"
    EULER = "euler"
    RUNGE_KUTTA = "rk4"
    STOCHASTIC = "stochastic"

class SolverArgs:
    """
    Options for ODE solver used by model
    """

    DEFAULT =  {
                "rtol": 1.4e-4,
                "atol": 1.4e-4,
            }
    PRECISE = {
                "rtol": 1.4e-8,
                "atol": 1.4e-8,
            }
    FAST =  {
                "rtol": 1.4e-3,
                "atol": 1.4e-3,
            }



def solve_ode(
    solver_type: str,
    ode_func: OdeFunction,
    values: np.ndarray,
    times: np.ndarray,
    solver_args: dict,
) -> np.ndarray:
    """
    Solve an ODE function given a function describing the dynamics, some initial conditions and times.
    """
    if solver_type == SolverType.ODE_INT:
        return solve_with_odeint(ode_func, values, times, solver_args)
    elif solver_type == SolverType.SOLVE_IVP:
        return solve_with_ivp(ode_func, values, times, solver_args)
    elif solver_type == SolverType.EULER:
        return solve_with_euler(ode_func, values, times, solver_args)
    elif solver_type == SolverType.RUNGE_KUTTA:
        return solve_with_rk4(ode_func, values, times, solver_args)
    else:
        raise ValueError("Solver type requested is not available")


def solve_with_odeint(
    ode_func: OdeFunction, values: np.ndarray, times: np.ndarray, solver_args: dict
):
    """
    Solve ODE with SciPy's odeint solver.

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.odeint.html
    """
    atol = solver_args.get("atol", 1e-3)
    rtol = solver_args.get("rtol", 1e-3)
    return odeint(ode_func, values, times, atol=atol, rtol=rtol)


DEFAULT_IVP_MAX_STEP = 1


def solve_with_ivp(ode_func: OdeFunction, values: np.ndarray, times: np.ndarray, solver_args: dict):
    """
    Solve ODE with SciPy's solve_ivp solver.
    This method allows us to set a stopping condition.

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html
    """

    def _ode_func(time, values):
        """Reverse parameters"""
        return ode_func(values, time)

    # Set default max step size to 1,
    # so that we do not skip over important time steps.
    if "max_step" not in solver_args:
        solver_args = {**solver_args, "max_step": DEFAULT_IVP_MAX_STEP}

    t_span = (times[0], times[-1])
    results = solve_ivp(_ode_func, t_span, values, t_eval=times, **solver_args)
    return results["y"].transpose()


def solve_with_euler(
    ode_func: OdeFunction, values: np.ndarray, times: np.ndarray, solver_args: dict
):
    """
    Solve ODE with a hand-rolled Euler's method implementation.

    `WARNING: This method is too inaccurate to use for real applications.`
    """
    step_size = solver_args.get("step_size", 0.1)
    start_time = times[0]
    end_time = times[-1]
    time_span = end_time - start_time
    num_timesteps = int(time_span / step_size) + 1
    assert (
        num_timesteps == time_span / step_size + 1
    ), f"Step size {step_size} must be a factor of the time span {time_span}."
    integration_times = np.linspace(start_time, end_time, num_timesteps)
    results_arr = np.zeros([num_timesteps, len(values)])
    results_arr[0] = np.array(values)

    # Perform Euler's method integration
    for time_idx, time in enumerate(integration_times[:-1]):
        values_arr = results_arr[time_idx]
        gradient_arr = ode_func(values_arr, time)
        results_arr[time_idx + 1] = values_arr + step_size * gradient_arr

    return _interpolate_solver_results(results_arr, integration_times, times)


def solve_with_rk4(ode_func: OdeFunction, values: np.ndarray, times: np.ndarray, solver_args: dict):
    """
    Solve ODE with a hand-rolled Runge-Kutta 4 implementation.

    `WARNING: This method may require a very small time step to perform accurate numerical integration.`
    """
    step_size = solver_args.get("step_size", 0.1)
    start_time = times[0]
    end_time = times[-1]
    time_span = end_time - start_time
    num_timesteps = int(time_span / step_size) + 1
    assert (
        num_timesteps == time_span / step_size + 1
    ), f"Step size {step_size} must be a factor of the time span {time_span}."
    integration_times = np.linspace(start_time, end_time, num_timesteps)
    results_arr = np.zeros([num_timesteps, len(values)])
    results_arr[0] = np.array(values)

    # Perform Runge-Kutta 4 method integration
    for time_idx, time in enumerate(integration_times[:-1]):
        values_arr = results_arr[time_idx]
        k1 = step_size * ode_func(values_arr, time)
        k2 = step_size * ode_func(values_arr + k1 / 2, time + step_size / 2)
        k3 = step_size * ode_func(values_arr + k2 / 2, time + step_size / 2)
        k4 = step_size * ode_func(values_arr + k3, time + step_size)
        results_arr[time_idx + 1] = values_arr + (1 / 6) * (k1 + 2 * k2 + 2 * k3 + k4)

    return _interpolate_solver_results(results_arr, integration_times, times)


def _interpolate_solver_results(results_arr, integration_times, requested_times):
    """
    Interpolate solver results into an output array that matches the requested times

    results_arr: Solver results, 2D Numpy array
    integration_times: Times used to get solver results
    requested_times: Times to interpolate
    """
    # Build a function to produce interpolated results
    solved_func = interp1d(integration_times, results_arr, axis=0)
    # Create output array to store values for requested times.
    output_arr = np.zeros([len(requested_times), results_arr.shape[1]])
    output_arr[0] = results_arr[0]
    # Populate output array with interpolated results
    num_times = len(requested_times)
    for time_idx in range(1, num_times):
        time = requested_times[time_idx]
        output_arr[time_idx] = solved_func(time)

    return output_arr



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['summer2',
 'summer2.compute',
 'summer2.experimental',
 'summer2.parameters',
 'summer2.runner',
 'summer2.runner.jax']

package_data = \
{'': ['*']}

install_requires = \
['computegraph>=0.3.2b2',
 'networkx>=2.6.2',
 'numpy>=1.20.3',
 'pandas>=1.3.2',
 'plotly>=5.5.0']

extras_require = \
{':sys_platform == "darwin"': ['jax[cpu]>=0.3.14,<0.4.0'],
 ':sys_platform == "linux"': ['jax[cpu]>=0.3.14,<0.4.0']}

setup_kwargs = {
    'name': 'summerepi2',
    'version': '1.0.1a7',
    'description': 'Summer is a compartmental disease modelling framework, written in Python. It provides a high-level API to build and run models.',
    'long_description': '# Summer: compartmental disease modelling in Python\n\n[![Automated Tests](https://github.com/monash-emu/summer/actions/workflows/tests.yml/badge.svg)](https://github.com/monash-emu/summer/actions/workflows/tests.yml)\n\nSummer is a Python-based framework for the creation and execution of [compartmental](https://en.wikipedia.org/wiki/Compartmental_models_in_epidemiology) (or "state-based") epidemiological models of infectious disease transmission.\n\nIt provides a range of structures for easily implementing compartmental models, including structure for some of the most common features added to basic compartmental frameworks, including:\n\n- A variety of inter-compartmental flows (infections, transitions, births, deaths, imports)\n- Force of infection multipliers (frequency, density)\n- Post-processing of compartment sizes into derived outputs\n- Stratification of compartments, including:\n  - Adjustments to flow rates based on strata\n  - Adjustments to infectiousness based on strata\n  - Heterogeneous mixing between strata\n  - Multiple disease strains\n\nSome helpful links to learn more:\n\n- [Rationale](http://summerepi.com/rationale.html) for why we are building Summer\n- **[Documentation](http://summerepi.com/)** with [code examples](http://summerepi.com/examples)\n- [Available on PyPi](https://pypi.org/project/summerepi2/) as `summerepi2`.\n- [Performance benchmarks](https://monash-emu.github.io/summer/)\n\n## Installation and Quickstart\n\nThis project requires at least Python 3.7\n\nSet up and activate an appropriate virtual environment, then install the `summerepi2` package from PyPI\n\n```bash\npip install summerepi2\n```\n\nImportant note for Windows users:\nsummerepi2 relies on the Jax framework for fast retargetable computing.  This is automatically\ninstalled under Linux, OSX, and WSL environments.  If you are using Windows, you can either install\nvia WSL, or run the following command after installing\n\n```bash\npip install jax[cpu]==0.3.14 -f https://whls.blob.core.windows.net/unstable/index.html\n```\n\nThen you can now use the library to build and run models. See [here](http://summerepi.com/examples) for some code examples.\n\n## Optional (recommended) extras\n\nSummer has advanced interactive plotting tools built in - but they are greatly improved with the\naddition of the pygraphviz library.\n\nIf you are using conda, the simplest method of installation is as follows:\n\n```bash\nconda install --channel conda-forge pygraphviz\n```\n\nFor other install methods, see\nhttps://pygraphviz.github.io/documentation/stable/install.html\n\n## Development\n\n[Poetry](https://python-poetry.org/) is used for packaging and dependency management.\n\nInitial project setup is documented [here](./docs/dev-setup.md) and should work for Windows or Ubuntu, maybe for MacOS.\n\nSome common things to do as a developer working on this codebase:\n\n```bash\n# Activate summer conda environment prior to doing other stuff (see setup docs)\nconda activate summer\n\n# Install latest requirements\npoetry install\n\n# Publish to PyPI - use your PyPI credentials\npoetry publish --build\n\n# Add a new package\npoetry add\n\n# Run tests\npytest -vv\n\n# Format Python code\nblack .\nisort . --profile black\n```\n\n## Releases\n\nReleases are numbered using [Semantic Versioning](https://semver.org/)\n\n- 1.0.0/1:\n  - Initial release\n- 1.1.0:\n  - Add stochastic integrator\n- 2.0.2:\n  - Rename fractional flow to transition flow\n  - Remove sojourn flow\n  - Add vectorized backend and other performance improvements\n- 2.0.3:\n  - Set default IVP solver to use a maximum step size of 1 timestep\n- 2.0.4:\n  - Add runtime derived values\n- 2.0.5:\n  - Remove legacy Summer implementation\n- 2.1.0:\n  - Add AdjustmentSystems\n  - Improve vectorization of flows\n  - Add computed_values inputs to flow and adjustment parameters\n- 2.1.1:\n  - Fix for invalid/unused package imports (cachetools)\n- 2.2.0\n  - Add validation and compartment caching optimizations\n- 2.2.1\n  - Derived output index caching\n  - Optimized fast-tracks for infectious multipliers\n- 2.2.2\n  - JIT infectiousness calculations\n  - Various micro-optimizations\n- 2.2.3\n  - Bugfix release (clamp outputs to 0.0)\n- 2.2.4\n  - Datetime awareness, DataFrame outputs\n- 2.2.5\n  - Performance improvements (frozenset), no API changes\n- 2.2.6\n  - Verify strata in flow adjustments (prevent unexpected behaviour)\n- 2.2.7\n  - Rename add_flow_adjustments -> set_flow_adjustments\n- 2.2.8\n  - Split imports functionality (add_importation_flow now requires split_imports arg)\n- 2.2.9\n  - Post-stratification population restribution\n- 2.3.0\n  - First official version to support only Python 3.7\n- 2.5.0\n  - Support Python 3.9\n- 2.6.0\n  - Merge 3.9/master branches\n- 2.7.0\n  - Include Python 3.10 support and update requirements\n- 2.7.1\n  - Bugfix (source flows not counted correctly in _add_transition_flow)\n- 3.1.0\n  - Parameter aware summer using computegraph\n- 3.1.1\n  - Update computegraph and add additional param awareness\n- 3.1.2\n  - Initial population/stratification now param aware\n- 3.1.3\n  - More jax implementation (strain stratification)\n- 3.1.4\n  - Bugfix release (jax imported even though optional)\n- 3.1.5\n  - Bugfix and minor improvements search for parameters\n- 3.1.6\n  - Support lists of adjustments and flow params\n- 4.0.0a\n  - Full jax support with ModelBuilder wrapper\n- 4.1.0a\n  - Abstract and lazy parameters\n- 4.2.0a\n  - Refactor parameters, combine runners\n\n## Release process\n\nTo do a release:\n\n- Commit any code changes and push them to GitHub\n- Choose a new release number accoridng to [Semantic Versioning](https://semver.org/)\n- Add a release note above\n- Edit the `version` key in `pyproject.toml` to reflect the release number\n- Publish the package to [PyPI](https://pypi.org/project/summerepi/) using Poetry, you will need a PyPI login and access to the project\n- Commit the release changes and push them to GitHub (Use a commit message like "Release 1.1.0")\n- Update `requirements.txt` in Autumn to use the new version of Summer\n\n```bash\npoetry build\npoetry publish\n```\n\n## Documentation\n\nSphinx is used to automatically build reference documentation for this library.\nThe documentation is automatically built and deployed to [summerepi.com](http://summerepi.com/) whenever code is pushed to `master`.\n\nTo run or edit the code examples in the documentation, start a jupyter notebook server as follows:\n\n```bash\njupyter notebook --config docs/jupyter_notebook_config.py\n# Go to http://localhost:8888/tree/docs/examples in your web browser.\n```\n\nYou can clean outputs from all the example notbooks with\n\n```bash\n./docs/scripts/clean.sh\n```\n\nTo build and deploy\n\n```bash\n./docs/scripts/build.sh\n./docs/scripts/deploy.sh\n```\n\nTo work on docs locally\n\n```bash\n./docs/scripts/watch.sh\n```\n',
    'author': 'James Trauer',
    'author_email': 'james.trauer@monash.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'http://summerepi.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""
This module contains classes which define adjustments to model parameters.
"""
from abc import ABC, abstractmethod
from typing import Callable, Union

from summer2.parameters import get_static_param_value
from summer2.parameters.param_impl import ModelParameter

FlowParam = Union[float, Callable[[float], float]]


class BaseAdjustment(ABC):
    """
    :meta private:
    """

    def __init__(self, param: FlowParam):
        self.param = param

    @abstractmethod
    def get_new_value(
        self, value: float, computed_values: dict, time: float, parameters: dict = None
    ) -> float:
        pass

    def _is_equal(self, adj):
        # Used for testing.
        return type(self) is type(adj) and self.param == adj.param

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.param}'>"

    def __hash__(self):
        return hash((type(self), self.param))

    def __eq__(self, o: object) -> bool:
        return self._is_equal(o)


class Multiply(BaseAdjustment):
    """
    A multiplication-based adjustment of a parameter.
    The new parameter value is the previous value, multiplied by the supplied parameter.

    Args:
        param: The parameter to be multiplied with the unadjusted value.


    Examples:
        Create an adjustment to multiply by a factor of 1.5::

            adjust.Multiply(1.5)

        Created an adjustment to multiply with the value of a time varying function::

            arbitrary_function = lambda time: 2 * time + 1
            adjust.Multiply(arbitrary_function)

    """

    def get_new_value(
        self, value: float, computed_values: dict, time: float, parameters: dict = None
    ) -> float:
        """
        Returns the adjusted value for a given time.

        Args:
            value: The value to adjust.
            time: The time to be used for any time-varying functions.

        Returns:
            float: The new, adjusted value.

        """
        # resolved_self = get_model_param_value(self.param, time, computed_values, parameters, True)
        if isinstance(self.param, ModelParameter):
            resolved_self = self.param.get_value(time, computed_values, parameters)
        elif callable(self.param):
            resolved_self = self.param(time, computed_values)
        else:
            resolved_self = self.param
        resolved_input = get_static_param_value(value, parameters, passthrough=True)
        return resolved_self * resolved_input


class Overwrite(BaseAdjustment):
    """
    Am overwrite-based adjustment of a parameter.
    The new parameter value is the supplied parameter, overwriting any previous values.

    Args:
        param: The parameter to be used in place of the unadjusted value.

    Examples:
        Create an adjustment to overwrite the previous value with 1.5::

            adjust.Overwrite(1.5)

        Create an adjustment to overwrite the previous value with the value of a
        time varying function::

            arbitrary_function = lambda time: 2 * time + 1
            adjust.Overwrite(arbitrary_function)

    """

    def get_new_value(
        self, value: float, computed_values: dict, time: float, parameters: dict = None
    ) -> float:
        """
        Returns the adjusted value for a given time.

        Args:
            value: The value to adjust.
            time: The time to be used for any time-varying functions.

        Returns:
            float: The new, adjusted value.

        """
        if isinstance(self.param, ModelParameter):
            return self.param.get_value(time, computed_values, parameters)
        elif callable(self.param):
            return self.param(time, computed_values)
        else:
            return self.param
        # return get_model_param_value(self.param, time, computed_values, parameters, True)


def enforce_wrapped(value, allowed, wrap):
    if any([isinstance(value, t) for t in allowed]):
        return value
    else:
        return wrap(value)


def enforce_multiply(value):
    return enforce_wrapped(value, [Multiply, Overwrite, type(None)], Multiply)



class ActionType:
    STRATIFY = "stratify"
    ADJUST_POP_SPLIT = "adjust_pop_split"


class TrackedAction:
    def __init__(self, action_type: ActionType, kwargs):
        self.action_type = action_type
        self.kwargs = kwargs


class ModelBuildTracker:
    def __init__(self):
        self.all_actions = []

    def append_action(self, action_type: ActionType, **kwargs):
        action = TrackedAction(action_type, kwargs)
        self.all_actions.append(action)
